from typing import Optional, List, Dict, Any
import gymnasium as gym
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_nmap_actions import EmulationAttackerNMAPActions
from csle_common.dao.emulation_action.attacker.emulation_attacker_nikto_actions import EmulationAttackerNIKTOActions
from csle_common.dao.emulation_action.attacker.emulation_attacker_masscan_actions \
    import EmulationAttackerMasscanActions
from csle_common.dao.emulation_action.attacker.emulation_attacker_network_service_actions \
    import EmulationAttackerNetworkServiceActions
from csle_common.dao.emulation_action.attacker.emulation_attacker_shell_actions import EmulationAttackerShellActions
from csle_common.dao.emulation_action.attacker.emulation_attacker_stopping_actions \
    import EmulationAttackerStoppingActions
from csle_base.json_serializable import JSONSerializable


class EmulationAttackerActionConfig(JSONSerializable):
    """
    Configuration of the action space for the attacker in the emulation
    """

    def __init__(self, num_indices: int, actions: Optional[List[EmulationAttackerAction]] = None,
                 nmap_action_ids: Optional[List[EmulationAttackerActionId]] = None,
                 network_service_action_ids: Optional[List[EmulationAttackerActionId]] = None,
                 shell_action_ids: Optional[List[EmulationAttackerActionId]] = None,
                 nikto_action_ids: Optional[List[EmulationAttackerActionId]] = None,
                 masscan_action_ids: Optional[List[EmulationAttackerActionId]] = None,
                 stopping_action_ids: Optional[List[EmulationAttackerActionId]] = None):
        """
        Class constructor

        :param num_indices: max num machine indexes allowed
        :param actions: list of actions in the action space
        :param nmap_action_ids: list of ids of the actions that are NMAP actions
        :param network_service_action_ids: list of ids of the actions that are network service actions
        :param shell_action_ids: list of ids of the actions that are shell actions
        :param nikto_action_ids: list of ids of the actions that are Nikto actions
        :param masscan_action_ids: list of ids of the actions that are Masscan actions
        :param stopping_action_ids: List of ids of the actions that are actions related to optimal stopping
        """
        if actions is None:
            raise ValueError("self.num_actions can not be defined")
        self.num_actions = len(actions)
        self.actions = actions
        self.num_indices = num_indices
        self.action_space = gym.spaces.Discrete(self.num_actions)
        self.action_lookup_d = {}
        self.action_lookup_d_val = {}
        for action in actions:
            self.action_lookup_d[(action.id, action.index)] = action
            self.action_lookup_d_val[(action.id, action.index)] = action

        self.nmap_action_ids = nmap_action_ids
        self.network_service_action_ids = network_service_action_ids
        self.shell_action_ids = shell_action_ids
        self.nikto_action_ids = nikto_action_ids
        self.masscan_action_ids = masscan_action_ids
        self.stopping_action_ids = stopping_action_ids
        if self.nmap_action_ids is None or self.network_service_action_ids is None or \
            self.shell_action_ids is None or self.nikto_action_ids is None or \
                self.masscan_action_ids is None or self.stopping_action_ids is None:
            raise ValueError("At least one of the actions is None")
        self.action_ids = (self.nmap_action_ids + self.network_service_action_ids + self.shell_action_ids
                           + self.nikto_action_ids + self.masscan_action_ids + self.stopping_action_ids)
        self.num_node_specific_actions = len(self.action_ids)
        self.m_action_space = gym.spaces.Discrete(self.num_node_specific_actions)
        self.ar_action_converter = {}

        # Add all (temp
        for j, a in enumerate(self.actions):
            idx2 = self.action_ids.index(a.id)
            key = (num_indices, idx2)
            self.ar_action_converter[key] = j

        # Add subnet actions
        for j, a in enumerate(self.actions):
            idx2 = self.action_ids.index(a.id)
            if a.index == self.num_indices:
                key = (num_indices, idx2)
                self.ar_action_converter[key] = j

        # Add subnet actions to all machines
        for i in range(num_indices):
            for j, a in enumerate(self.actions):
                idx2 = self.action_ids.index(a.id)
                if a.index == self.num_indices:
                    key = (i, idx2)
                    self.ar_action_converter[key] = j

        # Add rest of actions
        for i, a in enumerate(self.actions):
            if a.index != self.num_indices:
                idx2 = self.action_ids.index(a.id)
                key = (a.index, idx2)
                self.ar_action_converter[key] = i

    def print_actions(self) -> None:
        """
        Utility function for printing the list of actions

        :return: None
        """
        print("Attacker Actions:")
        for i, action in enumerate(self.actions):
            tag = "-"
            if not action.index == -1:
                if action.index is not None:
                    tag = str(action.index)
            else:
                tag = "*"
            print(str(i) + ":" + action.name + "[" + tag + "]")

    @staticmethod
    def dict_brute_same_user_ids():
        """
        :return: list of brute-force attack ids for teh attacker
        """
        return [
            EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL
        ]

    def get_continue_action_idx(self) -> int:
        """
        :return: the id of the continue-action
        """
        for i in range(len(self.actions)):
            if self.actions[i].id == EmulationAttackerActionId.CONTINUE:
                return i
        raise ValueError("No Continue Action in the action space")

    def get_action_by_id(self, action_id: EmulationAttackerActionId) -> EmulationAttackerAction:
        """
        Gets the action of a given id

        :param action_id: the action id
        :return: the action of the id
        """
        for a in self.actions:
            if a.id == action_id:
                return a
        raise ValueError("action id not found: {}".format(action_id))

    @staticmethod
    def all_actions_config(num_nodes: int, subnet_masks: List[str], hacker_ip: str) -> "EmulationAttackerActionConfig":
        """
        Gets the default action config for a given environment, which includes all actions

        :param num_nodes: the number of nodes in the environment
        :param subnet_masks: the subnet masks of the environment
        :param hacker_ip: the hacker ip
        :return:
        """
        attacker_actions = []

        # Host actions
        for idx in range(num_nodes):
            attacker_actions.append(EmulationAttackerNMAPActions.TCP_SYN_STEALTH_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.PING_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.UDP_PORT_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.TCP_CON_NON_STEALTH_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.TCP_FIN_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.TCP_NULL_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.TCP_XMAS_TREE_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.OS_DETECTION_SCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.NMAP_VULNERS(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.TELNET_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.SSH_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.FTP_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.CASSANDRA_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.IRC_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.MONGO_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.MYSQL_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.SMTP_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.POSTGRES_SAME_USER_PASS_DICTIONARY(index=idx))
            attacker_actions.append(EmulationAttackerNIKTOActions.NIKTO_WEB_HOST_SCAN(index=idx))
            attacker_actions.append(
                EmulationAttackerMasscanActions.MASSCAN_HOST_SCAN(index=idx, host_ip=hacker_ip))
            attacker_actions.append(EmulationAttackerNMAPActions.FIREWALK(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.HTTP_ENUM(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.HTTP_GREP(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.VULSCAN(index=idx))
            attacker_actions.append(EmulationAttackerNMAPActions.FINGER(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.SAMBACRY_EXPLOIT(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.SHELLSHOCK_EXPLOIT(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.DVWA_SQL_INJECTION(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.CVE_2015_3306_EXPLOIT(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.CVE_2015_1427_EXPLOIT(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.CVE_2016_10033_EXPLOIT(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.CVE_2010_0426_PRIV_ESC(index=idx))
            attacker_actions.append(EmulationAttackerShellActions.CVE_2015_5602_PRIV_ESC(index=idx))

        # Subnet actions
        attacker_actions.append(
            EmulationAttackerNMAPActions.TCP_SYN_STEALTH_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.PING_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.UDP_PORT_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(
            EmulationAttackerNMAPActions.TCP_CON_NON_STEALTH_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.TCP_FIN_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.TCP_NULL_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.TCP_XMAS_TREE_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.OS_DETECTION_SCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.NMAP_VULNERS(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.TELNET_SAME_USER_PASS_DICTIONARY(index=-1,
                                                                                              ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.SSH_SAME_USER_PASS_DICTIONARY(index=-1,
                                                                                           ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.FTP_SAME_USER_PASS_DICTIONARY(index=-1,
                                                                                           ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.CASSANDRA_SAME_USER_PASS_DICTIONARY(index=-1,
                                                                                                 ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.IRC_SAME_USER_PASS_DICTIONARY(index=-1,
                                                                                           ips=subnet_masks))
        attacker_actions.append(
            EmulationAttackerNMAPActions.MONGO_SAME_USER_PASS_DICTIONARY(index=-1, ips=subnet_masks))
        attacker_actions.append(
            EmulationAttackerNMAPActions.MYSQL_SAME_USER_PASS_DICTIONARY(index=-1, ips=subnet_masks))
        attacker_actions.append(
            EmulationAttackerNMAPActions.SMTP_SAME_USER_PASS_DICTIONARY(index=-1, ips=subnet_masks))
        attacker_actions.append(
            EmulationAttackerNMAPActions.POSTGRES_SAME_USER_PASS_DICTIONARY(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerShellActions.FIND_FLAG(index=-1))
        attacker_actions.append(EmulationAttackerNetworkServiceActions.SERVICE_LOGIN(index=-1))
        attacker_actions.append(EmulationAttackerMasscanActions.MASSCAN_HOST_SCAN(index=-1,
                                                                                  host_ip=hacker_ip, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.FIREWALK(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.HTTP_ENUM(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.HTTP_GREP(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.VULSCAN(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerNMAPActions.FINGER(index=-1, ips=subnet_masks))
        attacker_actions.append(EmulationAttackerShellActions.INSTALL_TOOLS(index=-1))
        attacker_actions.append(EmulationAttackerShellActions.SSH_BACKDOOR(index=-1))

        attacker_actions.append(EmulationAttackerStoppingActions.STOP(index=-1))
        attacker_actions.append(EmulationAttackerStoppingActions.CONTINUE(index=-1))

        attacker_actions = sorted(attacker_actions, key=lambda x: (x.id, x.index))
        nmap_action_ids = [
            EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_HOST,
            EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_ALL,
            EmulationAttackerActionId.PING_SCAN_HOST,
            EmulationAttackerActionId.PING_SCAN_ALL,
            EmulationAttackerActionId.UDP_PORT_SCAN_HOST,
            EmulationAttackerActionId.UDP_PORT_SCAN_ALL,
            EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_HOST,
            EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_ALL,
            EmulationAttackerActionId.TCP_FIN_SCAN_HOST,
            EmulationAttackerActionId.TCP_FIN_SCAN_ALL,
            EmulationAttackerActionId.TCP_NULL_SCAN_HOST,
            EmulationAttackerActionId.TCP_NULL_SCAN_ALL,
            EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_HOST,
            EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_ALL,
            EmulationAttackerActionId.OS_DETECTION_SCAN_HOST,
            EmulationAttackerActionId.OS_DETECTION_SCAN_ALL,
            EmulationAttackerActionId.NMAP_VULNERS_HOST,
            EmulationAttackerActionId.NMAP_VULNERS_ALL,
            EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST,
            EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL,
            EmulationAttackerActionId.FIREWALK_HOST, EmulationAttackerActionId.FIREWALK_ALL,
            EmulationAttackerActionId.HTTP_ENUM_HOST, EmulationAttackerActionId.HTTP_ENUM_ALL,
            EmulationAttackerActionId.HTTP_GREP_HOST, EmulationAttackerActionId.HTTP_GREP_ALL,
            EmulationAttackerActionId.VULSCAN_HOST, EmulationAttackerActionId.VULSCAN_ALL,
            EmulationAttackerActionId.FINGER_HOST, EmulationAttackerActionId.FINGER_ALL
        ]
        network_service_action_ids = [EmulationAttackerActionId.NETWORK_SERVICE_LOGIN]
        shell_action_ids = [EmulationAttackerActionId.FIND_FLAG,
                            EmulationAttackerActionId.INSTALL_TOOLS,
                            EmulationAttackerActionId.SSH_BACKDOOR,
                            EmulationAttackerActionId.SAMBACRY_EXPLOIT,
                            EmulationAttackerActionId.SHELLSHOCK_EXPLOIT,
                            EmulationAttackerActionId.DVWA_SQL_INJECTION,
                            EmulationAttackerActionId.CVE_2015_3306_EXPLOIT,
                            EmulationAttackerActionId.CVE_2015_1427_EXPLOIT,
                            EmulationAttackerActionId.CVE_2016_10033_EXPLOIT,
                            EmulationAttackerActionId.CVE_2010_0426_PRIV_ESC,
                            EmulationAttackerActionId.CVE_2015_5602_PRIV_ESC
                            ]
        nikto_action_ids = [EmulationAttackerActionId.NIKTO_WEB_HOST_SCAN]
        masscan_action_ids = [EmulationAttackerActionId.MASSCAN_HOST_SCAN, EmulationAttackerActionId.MASSCAN_ALL_SCAN]
        stopping_action_ids = [EmulationAttackerActionId.STOP, EmulationAttackerActionId.CONTINUE]
        attacker_action_config = EmulationAttackerActionConfig(num_indices=num_nodes + 1, actions=attacker_actions,
                                                               nmap_action_ids=nmap_action_ids,
                                                               network_service_action_ids=network_service_action_ids,
                                                               shell_action_ids=shell_action_ids,
                                                               nikto_action_ids=nikto_action_ids,
                                                               masscan_action_ids=masscan_action_ids,
                                                               stopping_action_ids=stopping_action_ids)
        return attacker_action_config

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"num_indices: {self.num_indices}, actions: {self.actions}, nmap_action_ids: {self.nmap_action_ids}, " \
               f"network_service_action_ids: {self.network_service_action_ids}, " \
               f"shell_action_ids: {self.shell_action_ids}, nikto_action_ids: {self.nikto_action_ids}, " \
               f"masscan_action_ids: {self.masscan_action_ids}, stopping_action_ids: {self.stopping_action_ids}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["num_indices"] = self.num_indices
        d["actions"] = self.actions
        d["nmap_action_ids"] = self.nmap_action_ids
        d["network_service_action_ids"] = self.network_service_action_ids
        d["shell_action_ids"] = self.shell_action_ids
        d["nikto_action_ids"] = self.nikto_action_ids
        d["masscan_action_ids"] = self.masscan_action_ids
        d["stopping_action_ids"] = self.stopping_action_ids
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationAttackerActionConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationAttackerActionConfig(
            num_indices=d["num_indices"], actions=d["actions"], nmap_action_ids=d["nmap_action_ids"],
            network_service_action_ids=d["network_service_action_ids"], shell_action_ids=d["shell_action_ids"],
            nikto_action_ids=d["nikto_action_ids"], masscan_action_ids=d["masscan_action_ids"],
            stopping_action_ids=d["stopping_action_ids"]
        )
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationAttackerActionConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationAttackerActionConfig.from_dict(json.loads(json_str))



from typing import Optional, List, Dict, Any
import time
from csle_common.dao.emulation_action.defender.emulation_defender_action_type import EmulationDefenderActionType
from csle_common.dao.emulation_action.defender.emulation_defender_action_id import EmulationDefenderActionId
from csle_common.dao.emulation_action.defender.emulation_defender_action_outcome import EmulationDefenderActionOutcome
from csle_base.json_serializable import JSONSerializable


class EmulationDefenderAction(JSONSerializable):
    """
    Class representing an action of the defender in the environment
    """

    def __init__(self, id: EmulationDefenderActionId, name: str, cmds: List[str],
                 type: EmulationDefenderActionType, descr: str,
                 ips: List[str], index: int,
                 action_outcome: EmulationDefenderActionOutcome = EmulationDefenderActionOutcome.GAME_END,
                 alt_cmds: Optional[List[str]] = None, execution_time: float = 0.0, ts: float = 0.0):
        """
        Class constructor

        :param id: id of the action
        :param name: name of the action
        :param cmds: command-line commands to apply the action on the emulation
        :param type: type of the action
        :param descr: description of the action (documentation)
        :param ips: ip of the machine to apply the action to
        :param index: index of the machine to apply the action to
        :param action_outcome: type of the outcome of the action
        :param alt_cmds: alternative command if the first command does not work
        :param execution_time: the time it took to run the action
        :param ts: the timestamp the action was run
        """
        self.id = id
        self.name = name
        self.cmds = cmds
        self.type = type
        self.descr = descr
        self.ips = ips
        self.action_outcome = action_outcome
        self.alt_cmds = alt_cmds
        if self.alt_cmds is None:
            self.alt_cmds = []
        self.index = index
        self.ts = ts
        self.execution_time = round(execution_time, 3)

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "id:{},name:{},ips:{},index:{}".format(self.id, self.name, self.ips, self.index)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationDefenderAction":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the instance
        """
        obj = EmulationDefenderAction(
            id=d["id"], name=d["name"], cmds=d["cmds"], type=d["type"], descr=d["descr"],
            ips=d["ips"],
            index=d["index"], action_outcome=d["action_outcome"], alt_cmds=d["alt_cmds"],
            ts=d["ts"], execution_time=d["execution_time"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dicr representation of the object
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["name"] = self.name
        d["cmds"] = list(self.cmds)
        d["type"] = self.type
        d["descr"] = self.descr
        d["ips"] = list(self.ips)
        d["index"] = self.index
        d["action_outcome"] = self.action_outcome
        d["alt_cmds"] = list(self.alt_cmds) if self.alt_cmds is not None else []
        d["execution_time"] = self.execution_time
        d["ts"] = self.ts
        return d

    def ips_match(self, ips: List[str]) -> bool:
        """
        Checks if a list of ips overlap with the ips of this host

        :param ips: the list of ips to check
        :return:  True if they match, False otherwise
        """
        for ip in self.ips:
            if ip in ips:
                return True
        return False

    def to_kafka_record(self) -> str:
        """
        Converts the instance into a kafka record format

        :param total_time: the total time of execution
        :return: the kafka record
        """
        ts = time.time()
        record = f"{ts},{self.id},{self.descr},{self.index},{self.name}," \
                 f"{self.execution_time},{'_'.join(self.ips)},{'_'.join(self.cmds)},{self.type}," \
                 f"{self.action_outcome},{'_'.join(self.alt_cmds) if self.alt_cmds is not None else []}"
        return record

    @staticmethod
    def from_kafka_record(record: str) -> "EmulationDefenderAction":
        """
        Converts a kafka record into an instance

        :param record: the record to convert
        :return: the created instance
        """
        parts = record.split(",")
        obj = EmulationDefenderAction(id=EmulationDefenderActionId(int(parts[1])), ts=float(parts[0]),
                                      descr=parts[2], index=int(parts[3]),
                                      name=parts[4],
                                      execution_time=float(parts[5]), ips=parts[6].split("_"),
                                      cmds=parts[7].split("_"),
                                      type=EmulationDefenderActionType(int(parts[8])),
                                      action_outcome=EmulationDefenderActionOutcome(int(parts[9])),
                                      alt_cmds=parts[10].split("_"))
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationDefenderAction":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationDefenderAction.from_dict(json.loads(json_str))

    def copy(self) -> "EmulationDefenderAction":
        """
        :return: a copy of the DTO
        """
        return EmulationDefenderAction.from_dict(self.to_dict())

    def num_attributes(self) -> int:
        """
        :return: The number of attributes of the DTO
        """
        return 11

    @staticmethod
    def schema() -> "EmulationDefenderAction":
        """
        :return: get the schema of the DTO
        """
        return EmulationDefenderAction(id=EmulationDefenderActionId.STOP, name="", cmds=[""],
                                       type=EmulationDefenderActionType.STOP, descr="", ips=[""], index=-1,
                                       action_outcome=EmulationDefenderActionOutcome.GAME_END, alt_cmds=[""],
                                       execution_time=0.0, ts=0.0)



from csle_common.dao.emulation_action_result.nikto_scan_result import NiktoScanResult
from csle_common.dao.emulation_action_result.nikto_vuln import NiktoVuln
from csle_common.dao.emulation_action_result.nmap_brute_credentials import NmapBruteCredentials
from csle_common.dao.emulation_action_result.nmap_hop import NmapHop
from csle_common.dao.emulation_action_result.nmap_http_enum import NmapHttpEnum
from csle_common.dao.emulation_action_result.nmap_http_grep import NmapHttpGrep
from csle_common.dao.emulation_action_result.nmap_os import NmapOs
from csle_common.dao.emulation_action_result.nmap_port import NmapPort
from csle_common.dao.emulation_action_result.nmap_vulscan import NmapVulscan
from csle_common.dao.emulation_action_result.nmap_trace import NmapTrace
from csle_common.dao.emulation_action_result.nmap_vuln import NmapVuln
from csle_common.dao.emulation_action_result.nmap_host_result import NmapHostResult
from csle_common.dao.emulation_action_result.nmap_scan_result import NmapScanResult


class TestEmulationActionResultDaoSuite:
    """
    Test suite for emulation action result data access objects (DAOs)
    """

    def test_nikto_vuln(self, example_nikto_vuln: NiktoVuln) -> None:
        """
        Tests creation and dict conversion of the NiktoVuln DAO

        :param example_nikto_vuln: an example NiktoVuln
        :return: None
        """
        assert isinstance(example_nikto_vuln.to_dict(), dict)
        assert isinstance(NiktoVuln.from_dict(example_nikto_vuln.to_dict()), NiktoVuln)
        assert NiktoVuln.from_dict(example_nikto_vuln.to_dict()).to_dict() == example_nikto_vuln.to_dict()
        assert NiktoVuln.from_dict(example_nikto_vuln.to_dict()) == example_nikto_vuln

    def test_nikto_scan_result(self, example_nikto_scan_result: NiktoScanResult) -> None:
        """
        Tests creation and dict conversion of the NiktoScanResult DAO

        :param example_nikto_scan_result: an example NiktoScanResult
        :return: None
        """
        assert isinstance(example_nikto_scan_result.to_dict(), dict)
        assert isinstance(NiktoScanResult.from_dict(example_nikto_scan_result.to_dict()), NiktoScanResult)
        assert NiktoScanResult.from_dict(example_nikto_scan_result.to_dict()).to_dict() == \
               example_nikto_scan_result.to_dict()
        assert NiktoScanResult.from_dict(example_nikto_scan_result.to_dict()) == example_nikto_scan_result

    def test_nmap_brute_credentials(self, example_nmap_brute_credentials: NmapBruteCredentials) -> None:
        """
        Tests creation and dict conversion of the NmapBruteCredentials DAO

        :param example_nmap_brute_credentials: an example NmapBruteCredentials
        :return: None
        """
        assert isinstance(example_nmap_brute_credentials.to_dict(), dict)
        assert isinstance(NmapBruteCredentials.from_dict(example_nmap_brute_credentials.to_dict()),
                          NmapBruteCredentials)
        assert NmapBruteCredentials.from_dict(example_nmap_brute_credentials.to_dict()).to_dict() \
               == example_nmap_brute_credentials.to_dict()
        assert NmapBruteCredentials.from_dict(example_nmap_brute_credentials.to_dict()) \
               == example_nmap_brute_credentials

    def test_nmap_hop(self, example_nmap_hop: NmapHop) -> None:
        """
        Tests creation and dict conversion of the NmapHop DAO

        :param example_nmap_hop: an example NmapHop
        :return: None
        """
        assert isinstance(example_nmap_hop.to_dict(), dict)
        assert isinstance(NmapHop.from_dict(example_nmap_hop.to_dict()), NmapHop)
        assert NmapHop.from_dict(example_nmap_hop.to_dict()).to_dict() == example_nmap_hop.to_dict()
        assert NmapHop.from_dict(example_nmap_hop.to_dict()) == example_nmap_hop

    def test_nmap_http_enum(self, example_nmap_http_enum: NmapHttpEnum) -> None:
        """
        Tests creation and dict conversion of the NmapHttpEnum DAO

        :param example_nmap_http_enum: an example NmapHttpEnum
        :return: None
        """
        assert isinstance(example_nmap_http_enum.to_dict(), dict)
        assert isinstance(NmapHttpEnum.from_dict(example_nmap_http_enum.to_dict()), NmapHttpEnum)
        assert NmapHttpEnum.from_dict(example_nmap_http_enum.to_dict()).to_dict() == example_nmap_http_enum.to_dict()
        assert NmapHttpEnum.from_dict(example_nmap_http_enum.to_dict()) == example_nmap_http_enum

    def test_nmap_http_grep(self, example_nmap_http_grep: NmapHttpGrep) -> None:
        """
        Tests creation and dict conversion of the NmapHTTPGrep DAO

        :param example_nmap_http_grep: an example NmapHttpGrep
        :return: None
        """
        assert isinstance(example_nmap_http_grep.to_dict(), dict)
        assert isinstance(NmapHttpGrep.from_dict(example_nmap_http_grep.to_dict()), NmapHttpGrep)
        assert NmapHttpGrep.from_dict(example_nmap_http_grep.to_dict()).to_dict() == example_nmap_http_grep.to_dict()
        assert NmapHttpGrep.from_dict(example_nmap_http_grep.to_dict()) == example_nmap_http_grep

    def test_nmap_vulscan(self, example_nmap_vulscan: NmapVulscan) -> None:
        """
        Tests creation and dict conversion of the NmapVulScan DAO

        :param example_nmap_vulscan: an example NmapVulscan
        :return: None
        """
        assert isinstance(example_nmap_vulscan.to_dict(), dict)
        assert isinstance(NmapVulscan.from_dict(example_nmap_vulscan.to_dict()), NmapVulscan)
        assert NmapVulscan.from_dict(example_nmap_vulscan.to_dict()).to_dict() == example_nmap_vulscan.to_dict()
        assert NmapVulscan.from_dict(example_nmap_vulscan.to_dict()) == example_nmap_vulscan

    def test_nmap_os(self, example_nmap_os: NmapOs) -> None:
        """
        Tests creation and dict conversion of the NmapHTTPGrep DAO

        :param example_nmap_os: an example NmapOs
        :return: None
        """
        assert isinstance(example_nmap_os.to_dict(), dict)
        assert isinstance(NmapOs.from_dict(example_nmap_os.to_dict()), NmapOs)
        assert NmapOs.from_dict(example_nmap_os.to_dict()).to_dict() == example_nmap_os.to_dict()
        assert NmapOs.from_dict(example_nmap_os.to_dict()) == example_nmap_os

    def test_nmap_port(self, example_nmap_port: NmapPort) -> None:
        """
        Tests creation and dict conversion of the NmapPort DAO

        :param example_nmap_port: an example NmapPort
        :return: None
        """
        assert isinstance(example_nmap_port.to_dict(), dict)
        assert isinstance(NmapPort.from_dict(example_nmap_port.to_dict()), NmapPort)
        assert NmapPort.from_dict(example_nmap_port.to_dict()).to_dict() == example_nmap_port.to_dict()
        assert NmapPort.from_dict(example_nmap_port.to_dict()) == example_nmap_port

    def test_nmap_trace(self, example_nmap_trace: NmapTrace) -> None:
        """
        Tests creation and dict conversion of the NmapTrace DAO

        :param example_nmap_trace: an example NmapTrace
        :return: None
        """
        assert isinstance(example_nmap_trace.to_dict(), dict)
        assert isinstance(NmapTrace.from_dict(example_nmap_trace.to_dict()), NmapTrace)
        assert NmapTrace.from_dict(example_nmap_trace.to_dict()).to_dict() == example_nmap_trace.to_dict()
        assert NmapTrace.from_dict(example_nmap_trace.to_dict()) == example_nmap_trace

    def test_nmap_vuln(self, example_nmap_vuln: NmapVuln) -> None:
        """
        Tests creation and dict conversion of the NmapVuln DAO

        :param example_nmap_vuln: an example NmapVuln
        :return: None
        """
        assert isinstance(example_nmap_vuln.to_dict(), dict)
        assert isinstance(NmapVuln.from_dict(example_nmap_vuln.to_dict()), NmapVuln)
        assert NmapVuln.from_dict(example_nmap_vuln.to_dict()).to_dict() == example_nmap_vuln.to_dict()
        assert NmapVuln.from_dict(example_nmap_vuln.to_dict()) == example_nmap_vuln

    def test_nmap_host_result(self, example_nmap_host_result: NmapHostResult) -> None:
        """
        Tests creation and dict conversion of the NmapHostResult

        :param example_nmap_host_result: an example NmapHostResult
        :return: None
        """
        assert isinstance(example_nmap_host_result.to_dict(), dict)
        assert isinstance(NmapHostResult.from_dict(example_nmap_host_result.to_dict()), NmapHostResult)
        assert NmapHostResult.from_dict(
            example_nmap_host_result.to_dict()).to_dict() == example_nmap_host_result.to_dict()
        assert NmapHostResult.from_dict(example_nmap_host_result.to_dict()) == example_nmap_host_result

    def test_nmap_scan_result(self, example_nmap_scan_result: NmapScanResult) -> None:
        """
        Tests creation and dict conversion of the NmapScanResult

        :return: None
        """
        assert isinstance(example_nmap_scan_result.to_dict(), dict)
        assert isinstance(NmapScanResult.from_dict(example_nmap_scan_result.to_dict()), NmapScanResult)
        assert NmapScanResult.from_dict(example_nmap_scan_result.to_dict()).to_dict() == \
               example_nmap_scan_result.to_dict()
        assert NmapScanResult.from_dict(example_nmap_scan_result.to_dict()) == example_nmap_scan_result



from typing import Dict, Any
from csle_common.dao.emulation_config.emulation_trace import EmulationTrace
from csle_common.dao.simulation_config.simulation_trace import SimulationTrace
from csle_base.json_serializable import JSONSerializable


class EmulationSimulationTrace(JSONSerializable):
    """
    DTO class representing a combined emulation and simulation trace
    """

    def __init__(self, emulation_trace: EmulationTrace, simulation_trace: SimulationTrace):
        """
        Intializes the DTO

        :param emulation_trace: the emulation trace
        :param simulation_trace: the simulation trace
        """
        self.emulation_trace = emulation_trace
        self.simulation_trace = simulation_trace
        self.id = -1

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationSimulationTrace":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationSimulationTrace(
            emulation_trace=EmulationTrace.from_dict(d["emulation_trace"]),
            simulation_trace=SimulationTrace.from_dict(d["simulation_trace"]),
        )
        obj.id = d["id"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["emulation_trace"] = self.emulation_trace.to_dict()
        d["simulation_trace"] = self.simulation_trace.to_dict()
        d["id"] = self.id
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"emulation_trace:{self.emulation_trace}, simulation_trace:{self.simulation_trace}, id:{self.id}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationSimulationTrace":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationSimulationTrace.from_dict(json.loads(json_str))



from typing import List
import threading
import time
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state import \
    EmulationDefenderMachineObservationState


class AvgHostMetricsThread(threading.Thread):
    """
    Thread that polls the Kafka log to get the latest status of the docker statistics for a specific host
    """

    def __init__(self, host_metrics: HostMetrics,
                 machines: List[EmulationDefenderMachineObservationState], sleep_time: int) -> None:
        """
        Initializes the thread

        :param host_metrics: the host metrics to update
        :param machines: the list of machiens to update the host metrics with
        """
        threading.Thread.__init__(self)
        self.machines = machines
        self.running = True
        self.host_metrics = host_metrics
        self.sleep_time = sleep_time

    def run(self) -> None:
        """
        Runs the thread

        :return: None
        """
        while self.running:
            time.sleep(self.sleep_time)
            total_num_logged_in_users = 0
            total_num_failed_login_attempts = 0
            total_num_open_connections = 0
            total_num_login_events = 0
            total_num_processes = 0
            total_num_users = 0
            for m in self.machines:
                if m.host_metrics is None:
                    raise ValueError("HostMetrics is None")
                total_num_logged_in_users += m.host_metrics.num_logged_in_users
                total_num_failed_login_attempts += m.host_metrics.num_failed_login_attempts
                total_num_open_connections += m.host_metrics.num_open_connections
                total_num_login_events += m.host_metrics.num_login_events
                total_num_processes += m.host_metrics.num_processes
                total_num_users += m.host_metrics.num_users
            avg_num_logged_in_users = int(total_num_logged_in_users / len(self.machines))
            avg_num_failed_login_attempts = int(total_num_failed_login_attempts / len(self.machines))
            avg_num_open_connections = int(total_num_open_connections / len(self.machines))
            avg_num_login_events = int(total_num_login_events / len(self.machines))
            avg_num_processes = int(total_num_processes / len(self.machines))
            avg_num_users = int(total_num_users / len(self.machines))
            self.host_metrics.num_logged_in_users = avg_num_logged_in_users
            self.host_metrics.num_failed_login_attempts = avg_num_failed_login_attempts
            self.host_metrics.num_open_connections = avg_num_open_connections
            self.host_metrics.num_login_events = avg_num_login_events
            self.host_metrics.num_processes = avg_num_processes
            self.host_metrics.num_users = avg_num_users



from typing import List, Dict, Any
import csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc
import csle_collector.docker_stats_manager.docker_stats_manager_pb2
import csle_collector.docker_stats_manager.docker_stats_util as docker_stats_util
from csle_base.json_serializable import JSONSerializable


class DockerStatsManagersInfo(JSONSerializable):
    """
    DTO containing the status of the docker stats managers for a given emulation execution
    """

    def __init__(self, ips: List[str], ports: List[int],
                 emulation_name: str, execution_id: int,
                 docker_stats_managers_statuses: List[
                     csle_collector.docker_stats_manager.docker_stats_manager_pb2.DockerStatsMonitorDTO],
                 docker_stats_managers_running: List[bool]):
        """
        Initializes the DTO

        :param docker_stats_managers_running: list of booleans that indicate whether hte docker stats manager
                are running
        :param ips: the list of IPs of the running docker stats managers
        :param ports: the list of ports of the running docker stats managers
        :param emulation_name: the name of the corresponding emulation
        :param execution_id: the ID of the corresponding emulation execution
        :param docker_stats_managers_statuses: a list of statuses of the Host managers
        """
        self.docker_stats_managers_running = docker_stats_managers_running
        self.ips = ips
        self.emulation_name = emulation_name
        self.execution_id = execution_id
        self.docker_stats_managers_statuses = docker_stats_managers_statuses
        self.ports = ports

    def __str__(self):
        """
        :return: a string representation of the DTO
        """
        return f"docker_stats_managers_running: {self.docker_stats_managers_running}, " \
               f"ips: {list(map(lambda x: str(x), self.ips))}, " \
               f"emulation_name: {self.emulation_name}, " \
               f"execution_id: {self.execution_id}, " \
               f"docker_stats_managers_statuses: {list(map(lambda x: str(x), self.docker_stats_managers_statuses))}," \
               f" ports: {list(map(lambda x: str(x), self.ports))}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["docker_stats_managers_running"] = self.docker_stats_managers_running
        d["ips"] = self.ips
        d["emulation_name"] = self.emulation_name
        d["execution_id"] = self.execution_id
        d["ports"] = self.ports
        d["docker_stats_managers_statuses"] = list(map(
            lambda x: docker_stats_util.DockerStatsUtil.docker_stats_monitor_dto_to_dict(x),
            self.docker_stats_managers_statuses))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DockerStatsManagersInfo":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = DockerStatsManagersInfo(
            docker_stats_managers_running=d["docker_stats_managers_running"], ips=d["ips"], ports=d["ports"],
            emulation_name=d["emulation_name"], execution_id=d["execution_id"],
            docker_stats_managers_statuses=list(map(
                lambda x: docker_stats_util.DockerStatsUtil.docker_stats_monitor_dto_from_dict(x),
                d["docker_stats_managers_statuses"])))
        return dto

    @staticmethod
    def from_json_file(json_file_path: str) -> "DockerStatsManagersInfo":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DockerStatsManagersInfo.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.dao.simulation_config.value_type import ValueType
from csle_common.dao.simulation_config.action import Action
from csle_base.json_serializable import JSONSerializable


class ActionSpaceConfig(JSONSerializable):
    """
    DTO Class representing the action space configuration of a player in a simulation environment
    """

    def __init__(self, actions: List[Action], player_id: int, action_type: ValueType, descr: str = ""):
        """
        Initializes the DTO

        :param actions: the list of actions
        :param action_type: the type of the actions
        :param descr: a description of the action space
        :param player_id: the id of the player
        """
        self.actions = actions
        self.action_type = action_type
        self.descr = descr
        self.player_id = player_id

    def actions_ids(self) -> List[int]:
        """
        :return: a list of action ids
        """
        return list(map(lambda x: x.id, self.actions))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ActionSpaceConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ActionSpaceConfig(
            actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
            action_type=d["action_type"], descr=d["descr"], player_id=d["player_id"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["action_type"] = self.action_type
        d["descr"] = self.descr
        d["player_id"] = self.player_id
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"actions:{list(map(lambda x: str(x), self.actions))}, action_type: {self.action_type}, " \
               f"descr: {self.descr}, player_id: {self.player_id}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ActionSpaceConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ActionSpaceConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_firewall_config import NodeFirewallConfig
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class TopologyConfig(JSONSerializable):
    """
    A DTO representing the topology configuration of an emulation environment
    """

    def __init__(self, node_configs: List[NodeFirewallConfig], subnetwork_masks: List[str]):
        """
        Initializes the DTO

        :param node_configs: the list of node configurations
        :param subnetwork: the subnetwork
        """
        self.node_configs = node_configs
        self.subnetwork_masks = subnetwork_masks

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TopologyConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = TopologyConfig(
            node_configs=list(map(lambda x: NodeFirewallConfig.from_dict(x), d["node_configs"])),
            subnetwork_masks=d["subnetwork_masks"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["subnetwork_masks"] = self.subnetwork_masks
        d["node_configs"] = list(map(lambda x: x.to_dict(), self.node_configs))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"node configs:{','.join(list(map(lambda x: str(x), self.node_configs)))}, " \
               f"subnetwork_masks:{','.join(self.subnetwork_masks)}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "TopologyConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return TopologyConfig.from_dict(json.loads(json_str))

    def copy(self) -> "TopologyConfig":
        """
        :return: a copy of the DTO
        """
        return TopologyConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "TopologyConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.subnetwork_masks = list(map(lambda x: GeneralUtil.replace_first_octet_of_ip(
            ip=x, ip_first_octet=ip_first_octet), config.subnetwork_masks))
        config.node_configs = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                       config.node_configs))
        return config



from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_config import EmulationAttackerActionConfig
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.defender.emulation_defender_action_config import EmulationDefenderActionConfig


class TestEmulationActionDaoSuite:
    """
    Test suite for emulation action data access objects (DAOs)
    """

    def test_emulation_attacker_action(self, example_emulation_attacker_action: EmulationAttackerAction) -> None:
        """
        Tests creation and dict conversion of the EmulationAttackerAction DAO

        :param example_emulation_attacker_action: an example EmulationAttackerAction
        :return: None
        """
        assert isinstance(example_emulation_attacker_action.to_dict(), dict)
        assert isinstance(EmulationAttackerAction.from_dict(example_emulation_attacker_action.to_dict()),
                          EmulationAttackerAction)
        assert EmulationAttackerAction.from_dict(example_emulation_attacker_action.to_dict()).to_dict() == \
               example_emulation_attacker_action.to_dict()
        assert EmulationAttackerAction.from_dict(example_emulation_attacker_action.to_dict()) \
               == example_emulation_attacker_action

    def test_emulation_attacker_action_config(
            self, example_emulation_attacker_action_config: EmulationAttackerActionConfig) -> None:
        """
        Tests creation and dict conversion of the EmulationAttackerActionConfig DAO

        :param example_emulation_attacker_action: an example EmulationAttackerActionConfig
        :return: None
        """
        assert isinstance(example_emulation_attacker_action_config.to_dict(), dict)
        assert isinstance(EmulationAttackerActionConfig.from_dict(example_emulation_attacker_action_config.to_dict()),
                          EmulationAttackerActionConfig)
        d1 = EmulationAttackerActionConfig.from_dict(example_emulation_attacker_action_config.to_dict()).to_dict()
        d2 = example_emulation_attacker_action_config.to_dict()
        assert d1 == d2
        assert EmulationAttackerActionConfig.from_dict(example_emulation_attacker_action_config.to_dict()) == \
               example_emulation_attacker_action_config

    def test_emulation_defender_action(self, example_emulation_defender_action: EmulationDefenderAction) -> None:
        """
        Tests creation and dict conversion of the EmulationDefenderAction DAO

        :param example_emulation_defender_action: an example EmulationDefenderAction
        :return: None
        """
        assert isinstance(example_emulation_defender_action.to_dict(), dict)
        assert isinstance(EmulationDefenderAction.from_dict(example_emulation_defender_action.to_dict()),
                          EmulationDefenderAction)
        assert EmulationDefenderAction.from_dict(example_emulation_defender_action.to_dict()).to_dict() == \
               example_emulation_defender_action.to_dict()
        assert EmulationDefenderAction.from_dict(example_emulation_defender_action.to_dict()) == \
               example_emulation_defender_action

    def test_emulation_defender_action_config(
            self, example_emulation_defender_action_config: EmulationDefenderActionConfig) -> None:
        """
        Tests creation and dict conversion of the EmulationDefenderActionConfig DAO

        :param example_emulation_defender_action_config: an example EmulationDefenderActionConfig
        :return: None
        """
        assert isinstance(example_emulation_defender_action_config.to_dict(), dict)
        assert isinstance(EmulationDefenderActionConfig.from_dict(example_emulation_defender_action_config.to_dict()),
                          EmulationDefenderActionConfig)
        d1 = EmulationDefenderActionConfig.from_dict(example_emulation_defender_action_config.to_dict()).to_dict()
        d2 = example_emulation_defender_action_config.to_dict()
        assert d1 == d2
        assert EmulationDefenderActionConfig.from_dict(example_emulation_defender_action_config.to_dict()) == \
               example_emulation_defender_action_config



import pytest
import logging
from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.ovs_controller import OVSController
from csle_common.dao.emulation_config.containers_config import ContainersConfig


class TestOVSControllerSuite:
    """
    Test suite for OVS controller
    """

    @pytest.fixture
    def mock_containers_config(self) -> ContainersConfig:
        """
        Test method that sets up OVS switches on containers

        :return: None
        """
        container1 = MagicMock()
        container1.physical_host_ip = "192.168.1.1"
        container1.name = "ovs_container1"
        container1.get_full_name.return_value = "container1"
        container1.ips_and_networks = [("192.168.1.100", MagicMock(interface="eth0", bitmask="255.255.255.0"))]
        container2 = MagicMock()
        container2.physical_host_ip = "192.168.1.2"
        container2.name = "ovs_container2"
        container2.get_full_name.return_value = "container2"
        container2.ips_and_networks = [("192.168.1.101", MagicMock(interface="eth0", bitmask="255.255.255.0"))]
        containers_config = MagicMock(spec=ContainersConfig)
        containers_config.containers = [container1, container2]
        return containers_config

    @patch("subprocess.Popen")
    @patch("time.sleep")
    def test_create_virtual_switches_on_container(self, mock_sleep, mock_popen, mock_containers_config) -> None:
        """
        Test method that creates the OVS switches

        :param mock_sleep: mock_sleep
        :param mock_popen: mock_popen
        :param mock_containers_config: mock_containers_config
        :return: None
        """
        logger = MagicMock()
        physical_server_ip = "192.168.1.1"
        constants.CONTAINER_IMAGES.OVS_IMAGES = ["ovs_container1"]
        OVSController.create_virtual_switches_on_container(mock_containers_config, physical_server_ip, logger)
        mock_popen.assert_called()
        assert mock_sleep.call_count == 5

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    def test_apply_ovs_config(self, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method that applies the OVS configuration

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        emulation_env_config = MagicMock()
        ovs_switch_config = MagicMock()
        ovs_switch_config.container_name = "ovs_container_1"
        ovs_switch_config.physical_host_ip = "192.168.1.1"
        ovs_switch_config.docker_gw_bridge_ip = "192.168.1.3"
        emulation_env_config.ovs_switch_config = ovs_switch_config
        emulation_env_config.ovs_switch_config.switch_configs = [ovs_switch_config]
        emulation_env_config.connections = {"192.168.1.3": MagicMock()}
        physical_server_ip = "192.168.1.2"
        logger = MagicMock(spec=logging.Logger)
        OVSController.apply_ovs_config(emulation_env_config=emulation_env_config, physical_server_ip=physical_server_ip,
                                       logger=logger)
        mock_connect_admin.assert_not_called()
        mock_execute_ssh_cmd.assert_not_called()



import logging
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil


class TopologyController:
    """
    Class managing topologies in the emulation environments
    """

    @staticmethod
    def create_topology(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                        logger: logging.Logger) -> None:
        """
        Utility function for connecting to a running emulation and creating the configuration

        :param emulation_env_config: the emulation configuration
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info("Creating topology")
        topology_configs = emulation_env_config.topology_config.node_configs
        topology_configs = topology_configs + [emulation_env_config.kafka_config.firewall_config,
                                               emulation_env_config.elk_config.firewall_config]
        if emulation_env_config.sdn_controller_config is not None:
            topology_configs = topology_configs + [emulation_env_config.sdn_controller_config.firewall_config]
        for node_fw_config in topology_configs:
            if node_fw_config.physical_host_ip != physical_server_ip:
                continue
            ip = node_fw_config.docker_gw_bridge_ip
            logger.info(f"Connecting to node:{ip}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            for route in node_fw_config.routes:
                target, gw = route
                cmd = f"{constants.COMMANDS.SUDO_ADD_ROUTE} {target} gw {gw}"
                logger.info(f"Adding route: {cmd} to routing table of node: {ip}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for default_network_fw_config in node_fw_config.ips_gw_default_policy_networks:
                if default_network_fw_config.default_gw is not None:
                    cmd = f"{constants.COMMANDS.SUDO_ADD_ROUTE} " \
                          f"-net {default_network_fw_config.network.subnet_mask.split('/')[0]} " \
                          f"{constants.COMMANDS.NETMASK} {default_network_fw_config.network.bitmask} " \
                          f"gw {default_network_fw_config.default_gw}"
                    logger.info(f"Adding default gw: {cmd} to routing table of node: {ip}")
                    EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                                  wait_for_completion=True)

            cmd = constants.COMMANDS.CLEAR_IPTABLES
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                          wait_for_completion=True)

            # Setup /etc/hosts
            internal_ip = node_fw_config.get_ips()[0]
            cmd = f"{constants.COMMANDS.ECHO} '" + internal_ip + " " + \
                  node_fw_config.hostname + f"' | {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} " \
                  f"{constants.ETC_HOSTS.DEFAULT_HOST_LINE_1} | {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} {constants.ETC_HOSTS.DEFAULT_HOST_LINE_2} " \
                  f"| {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} {constants.ETC_HOSTS.DEFAULT_HOST_LINE_3} " \
                  f"| {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} {constants.ETC_HOSTS.DEFAULT_HOST_LINE_4} " \
                  f"| {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} {constants.ETC_HOSTS.DEFAULT_HOST_LINE_5} " \
                  f"| {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            cmd = f"{constants.COMMANDS.ECHO} {constants.ETC_HOSTS.DEFAULT_HOST_LINE_6} " \
                  f"| {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            for node2 in emulation_env_config.topology_config.node_configs:
                ips2 = node_fw_config.get_ips()
                if ip not in ips2:
                    cmd = f"{constants.COMMANDS.ECHO} '" + ips2[0] + " " + node2.hostname \
                          + f"' | {constants.ETC_HOSTS.APPEND_TO_ETC_HOSTS}"
                    o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            # Setup iptables and arptables
            for output_node in node_fw_config.output_accept:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_OUTPUT} -d {output_node} -j {constants.FIREWALL.ACCEPT}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_OUTPUT} -d {output_node} -j {constants.FIREWALL.ACCEPT}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for input_node in node_fw_config.input_accept:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_INPUT} -s {input_node} -j {constants.FIREWALL.ACCEPT}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_INPUT} -s {input_node} -j {constants.FIREWALL.ACCEPT}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for forward_node in node_fw_config.forward_accept:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_FORWARD} -d {forward_node} -j {constants.FIREWALL.ACCEPT}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for output_node in node_fw_config.output_drop:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_OUTPUT} -d {output_node} -j {constants.FIREWALL.DROP}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_OUTPUT} -d {output_node} -j {constants.FIREWALL.DROP}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for input_node in node_fw_config.input_drop:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_INPUT} -s {input_node} -j {constants.FIREWALL.DROP}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_INPUT} -s {input_node} -j {constants.FIREWALL.DROP}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for forward_node in node_fw_config.forward_drop:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_FORWARD} -d {forward_node} -j {constants.FIREWALL.DROP}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            for default_network_fw_config in node_fw_config.ips_gw_default_policy_networks:
                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_OUTPUT} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_output}"
                logger.info(f"Adding firewall rule: {cmd}")
                o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                                        wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_OUTPUT} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_output}"
                logger.info(f"Adding firewall rule: {cmd}")
                o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                                        wait_for_completion=True)

                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_INPUT} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_input}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_INPUT} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_input}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

                cmd = f"{constants.COMMANDS.IPTABLES_APPEND_FORWARD} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_input}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)
                cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_FORWARD} -d " \
                      f"{default_network_fw_config.network.subnet_mask} -j " \
                      f"{default_network_fw_config.default_input}"
                logger.info(f"Adding firewall rule: {cmd}")
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                              wait_for_completion=True)

            # Default drops
            execution_subnet_mask = (f"{emulation_env_config.execution_id}.{emulation_env_config.level}"
                                     f"{constants.CSLE.CSLE_LEVEL_SUBNETMASK_SUFFIX}")
            cmd = f"{constants.COMMANDS.IPTABLES_APPEND_OUTPUT} -d " \
                  f"{execution_subnet_mask} -j " \
                  f"{constants.FIREWALL.DROP}"
            logger.info(f"Adding firewall rule: {cmd}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                          wait_for_completion=True)
            cmd = f"{constants.COMMANDS.IPTABLES_APPEND_FORWARD} -d " \
                  f"{execution_subnet_mask} -j " \
                  f"{constants.FIREWALL.DROP}"
            logger.info(f"Adding firewall rule: {cmd}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                          wait_for_completion=True)
            cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_OUTPUT} -d " \
                  f"{execution_subnet_mask} -j " \
                  f"{constants.FIREWALL.DROP}"
            logger.info(f"Adding firewall rule: {cmd}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                          wait_for_completion=True)
            cmd = f"{constants.COMMANDS.ARPTABLES_APPEND_FORWARD} -d " \
                  f"{execution_subnet_mask} -j " \
                  f"{constants.FIREWALL.DROP}"
            logger.info(f"Adding firewall rule: {cmd}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip),
                                          wait_for_completion=True)

            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)



import grpc
from unittest.mock import patch, MagicMock
from csle_common.util.grpc_util import GrpcUtil


class TestGrpcUtilSuite:
    """
    Test suite for grpc_util
    """

    @patch("grpc.channel_ready_future")
    def test_grpc_server_on(self, mock_channel_ready_future) -> None:
        """
        Test utility function to test if a given gRPC channel is working or not

        :param mock_channel_ready_future: mock_channel_ready_future
        :return: None
        """
        mock_future = MagicMock()
        mock_channel_ready_future.return_value = mock_future
        result = GrpcUtil.grpc_server_on(mock_channel_ready_future)
        mock_future.result.assert_called()
        assert result

    @patch("grpc.channel_ready_future")
    def test_grpc_server_on_timeout(self, mock_channel_ready_future) -> None:
        """
        Test utility function to test if a given gRPC channel is not working

        :param mock_channel_ready_future: mock_channel_ready_future
        :return: None
        """
        mock_future = MagicMock()
        mock_future.result.side_effect = grpc.FutureTimeoutError()
        mock_channel_ready_future.return_value = mock_future
        result = GrpcUtil.grpc_server_on(mock_channel_ready_future)
        mock_future.result.assert_called()
        assert not result



from typing import List, Dict, Any
from csle_common.dao.emulation_action_result.nmap_host_result import NmapHostResult
from csle_base.json_serializable import JSONSerializable


class NmapScanResult(JSONSerializable):
    """
    DTO representing the result of a NMAP Scan
    """

    def __init__(self, hosts: List[NmapHostResult], ips: List[str]):
        """
        Initializes the DTO

        :param hosts: the list of NMAP hosts
        :param ips: the ips of the host
        """
        self.hosts = hosts
        self.ips = ips
        self.reachable: List[str] = []
        for h in self.hosts:
            if h.ips is None:
                raise ValueError("No ips found")
            self.reachable = self.reachable + h.ips

    def __str__(self) -> str:
        """
        :return:  a string representation of the objectp
        """
        return "hosts:{}".format(" ".join(list(map(lambda x: str(x), self.hosts))))

    def copy(self) -> "NmapScanResult":
        """
        :return: a copy of the object
        """
        hosts = []
        for host in self.hosts:
            hosts.append(host.copy())
        return NmapScanResult(hosts=hosts, ips=self.ips)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["hosts"] = list(map(lambda x: x.to_dict(), self.hosts))
        d["ips"] = self.ips
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapScanResult":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapScanResult(hosts=list(map(lambda x: NmapHostResult.from_dict(x), d["hosts"])), ips=d["ips"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapScanResult":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapScanResult.from_dict(json.loads(json_str))



from typing import List, Tuple
import time


class SSHUtil:
    """
    Class containing utility functions for SSH connections
    """

    @staticmethod
    def execute_ssh_cmds(cmds: List[str], conn, wait_for_completion: bool = True) -> List[Tuple[bytes, bytes, float]]:
        """
        Executes a list of commands over an ssh connection to the emulation

        :param cmds: the list of commands
        :param conn: the ssh connection
        :param wait_for_completion: whether to wait for completion of the commands or not
        :return: List of tuples (outdata, errdata, total_time)
        """
        results = []
        for cmd in cmds:
            res = SSHUtil.execute_ssh_cmd(cmd=cmd, conn=conn, wait_for_completion=wait_for_completion)
            results.append(res)
        return results

    @staticmethod
    def execute_ssh_cmd(cmd: str, conn, wait_for_completion: bool = True, retries: int = 2) \
            -> Tuple[bytes, bytes, float]:
        """
        Executes an action on the emulation over a ssh connection,
        this is a synchronous operation that waits for the completion of the action before returning

        :param cmd: the command to execute
        :param conn: the ssh connection
        :param wait_for_completion: boolean flag whether to wait for completion or not
        :param retries: number of retries
        :return: outdata, errdata, total_time
        """
        exp = None
        for i in range(retries):
            try:
                transport_conn = conn.get_transport()
                session = transport_conn.open_session(timeout=128)
                start = time.time()
                session.exec_command(cmd)
                outdata, errdata = b'', b''
                # Wait for completion
                while True:
                    # Reading from output streams
                    while session.recv_ready():
                        outdata += session.recv(1000)
                    while session.recv_stderr_ready():
                        errdata += session.recv_stderr(1000)

                    # Check for completion
                    if session.exit_status_ready() or not wait_for_completion:
                        break
                end = time.time()
                total_time = end - start
                return outdata, errdata, total_time
            except Exception as e:
                exp = e
                time.sleep(10)
        raise ConnectionError(f"Connection failed: {str(exp)} {repr(exp)}")



"""
Type of vulnerabilities in an emulation
"""

from enum import IntEnum


class StaticEmulationAttackerType(IntEnum):
    """
    Enum representing the different static emulation attacker types
    """
    NOVICE = 0
    EXPERIENCED = 1
    EXPERT = 2



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class StatisticsDataset(JSONSerializable):
    """
    DTO representing a dataset of statistics
    """

    def __init__(self, name: str, description: str, download_count: int, file_path: str, url: str, date_added: float,
                 num_measurements: int, num_metrics: int, size_in_gb: float, compressed_size_in_gb: float,
                 citation: str, num_files: int, file_format: str, added_by: str, conditions: str, metrics: str,
                 num_conditions: int):
        """
        Initializes the DTO

        :param name: the name of the dataset
        :param description: the description of the dataset
        :param download_count: the download count
        :param file_path: the file path to the dataset
        :param url: the url to the dataset
        :param date_added: the date the dataset was added
        :param num_measurements: the number of measurements
        :param num_metrics: the number of metrics for collecting the measurements
        :param size_in_gb: the size in gb of the dataset (uncompressed)
        :param compressed_size_in_gb: the size of the dataset in gb (compresse)
        :param citation: the citation information
        :param num_files: the number of files
        :param added_by: the name of the person that added the dataset
        :param conditions: the list of conditions considered for the statistics
        :param metrics: the list of metrics included in the statistics
        :param num_conditions: the number of conditions the statistics were collected for
        """
        self.name = name
        self.description = description
        self.download_count = download_count
        self.file_path = file_path
        self.url = url
        self.date_added = date_added
        self.num_measurements = num_measurements
        self.num_metrics = num_metrics
        self.size_in_gb = size_in_gb
        self.compressed_size_in_gb = compressed_size_in_gb
        self.citation = citation
        self.num_files = num_files
        self.file_format = file_format
        self.added_by = added_by
        self.conditions = conditions
        self.metrics = metrics
        self.num_conditions = num_conditions
        self.id = -1

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["name"] = self.name
        d["description"] = self.description
        d["download_count"] = self.download_count
        d["file_path"] = self.file_path
        d["url"] = self.url
        d["date_added"] = self.date_added
        d["num_measurements"] = self.num_measurements
        d["num_metrics"] = self.num_metrics
        d["size_in_gb"] = self.size_in_gb
        d["compressed_size_in_gb"] = self.compressed_size_in_gb
        d["num_files"] = self.num_files
        d["citation"] = self.citation
        d["file_format"] = self.file_format
        d["added_by"] = self.added_by
        d["conditions"] = self.conditions
        d["metrics"] = self.metrics
        d["num_conditions"] = self.num_conditions
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "StatisticsDataset":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = StatisticsDataset(
            name=d["name"], description=d["description"], download_count=d["download_count"], file_path=d["file_path"],
            url=d["url"], date_added=d["date_added"], num_measurements=d["num_measurements"],
            num_metrics=d["num_metrics"], size_in_gb=d["size_in_gb"],
            compressed_size_in_gb=d["compressed_size_in_gb"], citation=d["citation"], num_files=d["num_files"],
            file_format=d["file_format"], added_by=d["added_by"], metrics=d["metrics"], conditions=d["conditions"],
            num_conditions=d["num_conditions"])
        if "id" in d:
            obj.id = d["id"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name: {self.name}, description: {self.description}, download_count: {self.download_count}, " \
               f"file_path: {self.file_path}, url: {self.url}, date_added: {self.date_added}, " \
               f"num_measurements: {self.num_measurements}, " \
               f"num_metrisc: {self.num_metrics}," \
               f"size_in_gb: {self.size_in_gb}, compressed_size_in_gb: {self.compressed_size_in_gb}, " \
               f"citation: {self.citation}, num_files: {self.num_files}, " \
               f"file_format: {self.file_format}, added_by: {self.added_by}, metrics: {self.metrics}, " \
               f"conditions: {self.conditions}, num_conditions: {self.num_conditions}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "StatisticsDataset":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return StatisticsDataset.from_dict(json.loads(json_str))



"""
Utility functions for training with the csle environments
"""
from typing import List
import io
import json
import logging
import time
import argparse
import os
import sys
import numpy as np
import random
import torch
import scipy.stats
from scipy.interpolate import interp1d
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.encoding.np_encoder import NpEncoder
import csle_common.constants.constants as constants
from typing import Optional


class ExperimentUtil:
    """
    Class with utility functions related to training
    """

    @staticmethod
    def get_subdir(output_dir: str, results_dir: str, subdir: str, seed: int) -> str:
        """
        Utility function to construct the subdir string from a given results dir, subdir, and random seed

        :param output_dir: the base output directory
        :param results_dir: the base results dir
        :param subdir: the subdirectory (e.g. logs, data, tensorboard, etc)
        :param seed: the random seed
        :return: the directory path
        """
        return (output_dir + constants.COMMANDS.SLASH_DELIM + results_dir + constants.COMMANDS.SLASH_DELIM + subdir
                + constants.COMMANDS.SLASH_DELIM + str(seed) + constants.COMMANDS.SLASH_DELIM)

    @staticmethod
    def create_artifact_dirs(output_dir: str, random_seed: int) -> None:
        """
        Creates artifact directories if they do not already exist

        :param output_dir: the base directory
        :param random_seed: the random seed of the experiment
        :return: None
        """
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.LOG_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.LOG_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.PLOTS_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.PLOTS_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.DATA_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.DATA_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.HYPERPARAMETERS_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.HYPERPARAMETERS_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.GIFS_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.GIFS_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.TENSORBOARD_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.TENSORBOARD_DIR, random_seed))
        if not os.path.exists(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                        constants.EXPERIMENT.ENV_DATA_DIR, random_seed)):
            os.makedirs(ExperimentUtil.get_subdir(output_dir, constants.EXPERIMENT.RESULTS_DIR,
                                                  constants.EXPERIMENT.ENV_DATA_DIR, random_seed))

    @staticmethod
    def setup_experiment_logger(name: str, logdir: str, time_str: Optional[str] = None):
        """
        Configures the logger for writing log-data of training

        :param name: name of the logger
        :param logdir: directory to save log files
        :param time_str: time string for file names
        :return: None
        """
        # create formatter
        formatter = logging.Formatter('%(asctime)s,%(message)s')
        # log to console
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)

        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        # log to file
        if time_str is None:
            time_str = str(time.time())
        fh = logging.FileHandler(logdir + constants.COMMANDS.SLASH_DELIM + time_str +
                                 constants.COMMANDS.UNDERSCORE_DELIM +
                                 name + constants.FILE_PATTERNS.LOG_SUFFIX, mode="w")
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)

        # add the handlers to the logger
        logger.addHandler(fh)
        return logger

    @staticmethod
    def write_emulation_config_file(emulation_env_config: EmulationEnvConfig, path: str) -> None:
        """
        Writes a config object to a config file

        :param emulation_env_config: the emulation env config object
        :param path: the path to write the file
        :return: None
        """
        json_str = json.dumps(emulation_env_config.to_dict(), indent=4, sort_keys=True)
        with io.open(path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def write_simulation_config_file(simulation_env_config: SimulationEnvConfig, path: str) -> None:
        """
        Writes a config object to a config file

        :param simulation_env_config: the simulation env config object
        :param path: the path to write the file
        :return: None
        """
        json_str = json.dumps(simulation_env_config.to_dict(), indent=4, sort_keys=True, cls=NpEncoder)
        with io.open(path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def read_emulation_env_config(emulation_env_config_path) -> EmulationEnvConfig:
        """
        Reads emulation env config from a json file

        :param emulation_env_config_path: the path to the emulation env config file
        :return: the emulation env configuration
        """
        with io.open(emulation_env_config_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
        emulation_env_config: EmulationEnvConfig = EmulationEnvConfig.from_dict(json.loads(json_str))
        return emulation_env_config

    @staticmethod
    def read_simulation_env_config(simulation_env_config_path) -> SimulationEnvConfig:
        """
        Reads simulation env config from a json file

        :param simulation_env_config_path: the path to the simulation env config file
        :return: the simulation env configuration
        """
        with io.open(simulation_env_config_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
        simulation_env_config: SimulationEnvConfig = SimulationEnvConfig.from_dict(json.loads(json_str))
        return simulation_env_config

    @staticmethod
    def read_env_picture(env_picture_path) -> bytes:
        """
        Reads the environment topology picture from a file

        :param env_picture_path: the path to picture
        :return: the emulation env configuration
        """
        with open(env_picture_path, "rb") as image_file:
            image_data = image_file.read()
            return image_data

    @staticmethod
    def read_containers_config(containers_config_path) -> ContainersConfig:
        """
        Reads containers_config of the experiment from a json file

        :param containers_config_path: the path to the containers_config file
        :return: the configuration
        """
        with io.open(containers_config_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
        containers_cfg: ContainersConfig = ContainersConfig.from_dict(json.loads(json_str))
        return containers_cfg

    @staticmethod
    def parse_args(default_config_path):
        """
        Parses the commandline arguments with argparse

        :param default_config_path: default path to config file
        """
        parser = argparse.ArgumentParser(description='Parse flags to configure the json parsing')
        parser.add_argument("-cp", "--configpath", help="Path to configuration file",
                            default=default_config_path, type=str)
        parser.add_argument("-po", "--plotonly", help="Boolean parameter, if true, only plot",
                            action="store_true")
        parser.add_argument("-nc", "--noconfig", help="Boolean parameter, if true always override config",
                            action="store_true")
        parser.add_argument("-cs", "--csvfile", help="CSV file for plotting", type=str)
        parser.add_argument("-rd", "--resultdirs", help="List of comma-separated result dirs to combine for plotting",
                            type=str)
        args = parser.parse_args()
        return args

    @staticmethod
    def get_script_path():
        """
        :return: the script path
        """
        return os.path.dirname(os.path.realpath(sys.argv[0]))

    @staticmethod
    def default_output_dir() -> str:
        """
        :return: the default output dir
        """
        script_dir = ExperimentUtil.get_script_path()
        return str(script_dir)

    @staticmethod
    def default_emulation_config_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to emulation config file
        """
        if out_dir is None:
            config_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                       + constants.DOCKER.EMULATION_ENV_CFG_PATH)
        else:
            config_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM
                                       + constants.DOCKER.EMULATION_ENV_CFG_PATH)
        return config_path

    @staticmethod
    def default_simulation_config_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to simulatio config file
        """
        if out_dir is None:
            config_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                       + constants.SIMULATION.SIMULATION_ENV_CFG_PATH)
        else:
            config_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM
                                       + constants.SIMULATION.SIMULATION_ENV_CFG_PATH)
        return config_path

    @staticmethod
    def default_emulation_picture_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to emulation img file
        """
        if out_dir is None:
            img_path = os.path.join(ExperimentUtil.default_output_dir(), (constants.COMMANDS.DOT_DELIM +
                                                                          constants.DOCKER.EMULATION_ENV_IMAGE))
        else:
            img_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM + constants.DOCKER.EMULATION_ENV_IMAGE)
        return img_path

    @staticmethod
    def default_simulation_picture_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to simulatio img file
        """
        if out_dir is None:
            img_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                    + constants.DOCKER.SIMULATION_ENV_IMAGE)
        else:
            img_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM
                                    + constants.DOCKER.SIMULATION_ENV_IMAGE)
        return img_path

    @staticmethod
    def default_containers_folders_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to container folders
        """
        if out_dir is None:
            config_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                       + constants.COMMANDS.SLASH_DELIM
                                       + constants.DOCKER.CONTAINERS_DIR)
        else:
            config_path = os.path.join(out_dir, constants.DOCKER.CONTAINERS_DIR)
        return config_path

    @staticmethod
    def default_makefile_template_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to makefile tempalte
        """
        if out_dir is None:
            config_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                       + constants.COMMANDS.SLASH_DELIM
                                       + constants.DOCKER.MAKEFILE_TEMPLATE)
        else:
            config_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM
                                       + constants.COMMANDS.SLASH_DELIM + constants.DOCKER.MAKEFILE_TEMPLATE)
        return config_path

    @staticmethod
    def default_makefile_path(out_dir: Optional[str] = None) -> str:
        """
        :param out_dir: directory to write
        :return: the default path to makefile tempalte
        """
        if out_dir is None:
            config_path = os.path.join(ExperimentUtil.default_output_dir(), constants.COMMANDS.DOT_DELIM
                                       + constants.COMMANDS.SLASH_DELIM +
                                       constants.DOCKER.MAKEFILE)
        else:
            config_path = os.path.join(out_dir, constants.COMMANDS.DOT_DELIM
                                       + constants.COMMANDS.SLASH_DELIM + constants.DOCKER.MAKEFILE)
        return config_path

    @staticmethod
    def running_average(x, N):
        """
        Function used to compute the running average of the last N elements of a vector x

        :param x: the vector to take the running average
        :param N: the lenght of the running average
        :return: the averaged vector
        """
        if len(x) < N:
            N = len(x)
        if len(x) >= N:
            y = np.copy(x)
            y[N - 1:] = np.convolve(x, np.ones((N, )) / N, mode='valid')
        else:
            y = np.zeros_like(x)
        return round(y[-1], 2)

    @staticmethod
    def running_average_list(x, N):
        """
        Function used to compute the running average of the last N elements of a vector x

        :param x: the vector to take the running average of
        :param N: the running average length
        :return: a list with running averages
        """
        if len(x) < N:
            N = len(x)
        if len(x) >= N:
            y = np.copy(x)
            y[N - 1:] = np.convolve(x, np.ones((N, )) / N, mode='valid')
        else:
            y = np.zeros_like(x)
        return y

    @staticmethod
    def mean_confidence_interval(data, confidence=0.95):
        """
        Compute confidence intervals

        :param data: the data
        :param confidence: the interval confidence
        :return: the mean, the lower confidence interval, the upper confidence interval
        """
        a = 1.0 * np.array(data)
        n = len(a)
        m, se = np.mean(a), scipy.stats.sem(a)
        h = se * scipy.stats.t.ppf((1 + confidence) / 2., n - 1)
        return m, h

    @staticmethod
    def set_seed(seed: int) -> None:
        """
        Sets the random seed

        :param seed: the seed to set
        :return: None
        """
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

    @staticmethod
    def regress_lists(lists: List[List[float]]) -> List[List[float]]:
        """
        Regress sublists by interpolating their values to fit the length of the first sublist (reference list).

        :param lists: the lists ro regress
        :return: the regressed lists
        """
        # Assume that the longest list is the reference list
        ref_len = max([len(sublist) for sublist in lists])
        for i in range(len(lists)):
            if len(lists[i]) != ref_len:
                # If the list is not the same length, interpolate to the correct size
                sublist = lists[i]
                x = np.linspace(0, 1, len(sublist))
                y = np.array(sublist).squeeze()
                f = interp1d(x, y, kind="linear")
                new_x = np.linspace(0, 1, ref_len)
                new_y = f(new_x)
                lists[i] = new_y
        return lists



from typing import List
import grpc
import time
import logging
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.kafka_managers_info import KafkaManagersInfo
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
import csle_collector.kafka_manager.kafka_manager_pb2_grpc
import csle_collector.kafka_manager.kafka_manager_pb2
import csle_collector.kafka_manager.query_kafka_server
import csle_collector.kafka_manager.kafka_manager_util
from csle_common.util.emulation_util import EmulationUtil
from csle_common.logging.log import Logger


class KafkaController:
    """
    Controller for kafka, implements logic for managing kafka instances.
    """

    @staticmethod
    def start_kafka_manager(emulation_env_config: EmulationEnvConfig) -> None:
        """
        Utility method for starting the Kafka manager

        :param emulation_env_config: the emulation env config
        :return: None
        """
        # Check if kafka_manager is already running
        status = None
        not_running = False
        try:
            status = KafkaController.get_kafka_status_by_port_and_ip(
                ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                port=emulation_env_config.kafka_config.kafka_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = f"running: {status.running}, topics: {status.topics}"
        if not_running:
            Logger.__call__().get_logger().info(f"Starting the Kafka manager on node "
                                                f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}")
            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                        ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                        create_producer=False)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.KAFKA_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip))

            # Start the kafka_manager
            cmd = constants.COMMANDS.START_KAFKA_MANAGER.format(
                emulation_env_config.kafka_config.kafka_manager_port,
                emulation_env_config.kafka_config.kafka_manager_log_dir,
                emulation_env_config.kafka_config.kafka_manager_log_file,
                emulation_env_config.kafka_config.kafka_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip))
            time.sleep(2)
        else:
            Logger.__call__().get_logger().info(f"The Kafka manager is already running on node "
                                                f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}. "
                                                f"Status: {status_str}")

    @staticmethod
    def stop_kafka_manager(emulation_env_config: EmulationEnvConfig) -> None:
        """
        Utility method for stopping the Kafka manager

        :param emulation_env_config: the emulation env config
        :return: None
        """

        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                    create_producer=False)

        Logger.__call__().get_logger().info(f"Stopping the Kafka manager on node "
                                            f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}")

        # Stop old background job if running
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.KAFKA_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(
            cmd=cmd,
            conn=emulation_env_config.get_connection(
                ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip))
        time.sleep(2)

    @staticmethod
    def create_topics(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        A method that sends a request to the KafkaManager to create topics according to the given configuration

        :param emulation_env_config: the configuration of the emulation env
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(
            f"Creating kafka topics on container: {emulation_env_config.kafka_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.kafka_config.container.get_ips()[0]}, "
            f"{emulation_env_config.kafka_config.container.get_full_name()})")
        KafkaController.start_kafka_manager(emulation_env_config=emulation_env_config)
        kafka_dto = KafkaController.get_kafka_status_by_port_and_ip(
            ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
            port=emulation_env_config.kafka_config.kafka_manager_port)
        with grpc.insecure_channel(
                f'{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.kafka_config.kafka_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub(channel)

            if not kafka_dto.running:
                # Open a gRPC session
                logger.info("Kafka server is not running, starting it.")

                csle_collector.kafka_manager.query_kafka_server.start_kafka(stub)
                time.sleep(20)

            for topic in emulation_env_config.kafka_config.topics:
                logger.info(f"Creating topic: {topic.name}")
                csle_collector.kafka_manager.query_kafka_server.create_topic(
                    stub, name=topic.name, partitions=topic.num_partitions, replicas=topic.num_replicas,
                    retention_time_hours=topic.retention_time_hours
                )

    @staticmethod
    def get_kafka_status(emulation_env_config: EmulationEnvConfig) -> \
            csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO:
        """
        Method for querying the KafkaManager about the status of the Kafka server

        :param emulation_env_config: the emulation config
        :return: a KafkaDTO with the status of the server
        """
        KafkaController.start_kafka_manager(emulation_env_config=emulation_env_config)
        kafka_dto = KafkaController.get_kafka_status_by_port_and_ip(
            ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
            port=emulation_env_config.kafka_config.kafka_manager_port)
        return kafka_dto

    @staticmethod
    def get_kafka_status_by_port_and_ip(
            ip: str, port: int, timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) -> (
            csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO):
        """
        Method for querying the KafkaManager about the status of the Kafka server

        :param ip: the ip where the KafkaManager is running
        :param port: the port the KafkaManager is listening to
        :param timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS
        :return: a KafkaDTO with the status of the server
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub(channel)
            kafka_dto = csle_collector.kafka_manager.query_kafka_server.get_kafka_status(stub, timeout=timeout)
            return kafka_dto

    @staticmethod
    def stop_kafka_server(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO:
        """
        Method for requesting the KafkaManager to stop the Kafka server

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: a KafkaDTO with the status of the server
        """
        logger.info(
            f"Stopping Kafka server on container: {emulation_env_config.kafka_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.kafka_config.container.get_ips()[0]}, "
            f"{emulation_env_config.kafka_config.container.get_full_name()})")
        KafkaController.start_kafka_manager(emulation_env_config=emulation_env_config)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.kafka_config.kafka_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub(channel)
            kafka_dto = csle_collector.kafka_manager.query_kafka_server.stop_kafka(stub)
            return kafka_dto

    @staticmethod
    def configure_broker_ips(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Method for configuring the broker IPs on the Kafka container

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: a KafkaDTO with the status of the server
        """
        logger.info(
            f"Configuring broker IPs on container: {emulation_env_config.kafka_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.kafka_config.container.get_ips()[0]}, "
            f"{emulation_env_config.kafka_config.container.get_full_name()})")

        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip)
        sftp_client = emulation_env_config.get_connection(
            ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip).open_sftp()
        remote_file = sftp_client.open(collector_constants.KAFKA.KAFKA_CONFIG_FILE, mode="r")
        try:
            file_contents: str = remote_file.read().decode()
            file_contents = file_contents.replace(collector_constants.KAFKA.INTERNAL_IP_PLACEHOLDER,
                                                  emulation_env_config.kafka_config.container.get_ips()[0])
            file_contents = file_contents.replace(collector_constants.KAFKA.EXTERNAL_IP_PLACEHOLDER,
                                                  emulation_env_config.kafka_config.container.docker_gw_bridge_ip)
        finally:
            remote_file.close()

        remote_file = sftp_client.open(collector_constants.KAFKA.KAFKA_CONFIG_FILE, mode="w")
        try:
            remote_file.write(data=file_contents)
            remote_file.flush()
        finally:
            remote_file.close()

    @staticmethod
    def start_kafka_server(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.kafka_manager.kafka_manager_pb2.KafkaDTO:
        """
        Method for requesting the KafkaManager to start the Kafka server

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: a KafkaDTO with the status of the server
        """
        logger.info(
            f"Starting Kafka server on container: {emulation_env_config.kafka_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.kafka_config.container.get_ips()[0]}, "
            f"{emulation_env_config.kafka_config.container.get_full_name()})")
        KafkaController.start_kafka_manager(emulation_env_config=emulation_env_config)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.kafka_config.kafka_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.kafka_manager.kafka_manager_pb2_grpc.KafkaManagerStub(channel)
            kafka_dto = csle_collector.kafka_manager.query_kafka_server.start_kafka(stub)
            return kafka_dto

    @staticmethod
    def get_kafka_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the Kafka managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [emulation_env_config.kafka_config.container.docker_gw_bridge_ip]

    @staticmethod
    def get_kafka_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the IPS of the Kafka managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [emulation_env_config.kafka_config.kafka_manager_port]

    @staticmethod
    def get_kafka_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                logger: logging.Logger, physical_host_ip: str) -> KafkaManagersInfo:
        """
        Extracts the information of the Kafka managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param logger: the logger to use for logging
        :param physical_host_ip: the IP of the physical host
        :return: a DTO with the status of the Kafka managers
        """
        kafka_managers_ips = KafkaController.get_kafka_managers_ips(emulation_env_config=emulation_env_config)
        kafka_managers_ports = KafkaController.get_kafka_managers_ports(emulation_env_config=emulation_env_config)
        kafka_managers_statuses = []
        kafka_managers_running = []
        if physical_host_ip == emulation_env_config.kafka_config.container.physical_host_ip:
            for ip in kafka_managers_ips:
                if ip not in active_ips:
                    continue
                running = False
                status = None
                try:
                    status = KafkaController.get_kafka_status_by_port_and_ip(
                        port=emulation_env_config.kafka_config.kafka_manager_port, ip=ip)
                    running = True
                except Exception as e:
                    logger.debug(
                        f"Could not fetch Kafka manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
                if status is not None:
                    kafka_managers_statuses.append(status)
                else:
                    kafka_managers_statuses.append(
                        csle_collector.kafka_manager.kafka_manager_util.KafkaManagerUtil.kafka_dto_empty())
                kafka_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        kafka_manager_info_dto = KafkaManagersInfo(
            kafka_managers_running=kafka_managers_running, ips=kafka_managers_ips, execution_id=execution_id,
            emulation_name=emulation_name, kafka_managers_statuses=kafka_managers_statuses,
            ports=kafka_managers_ports)
        return kafka_manager_info_dto



from typing import Dict, Any

import numpy as np
from numpy.typing import NDArray

from csle_base.json_serializable import JSONSerializable


class RewardFunctionConfig(JSONSerializable):
    """
    DTO containing the reward tensor of a simulation
    """

    def __init__(self, reward_tensor: NDArray[Any]):
        """
        Initalizes the DTO

        :param reward_tensor: the reward tensor of the simulation
        """
        self.reward_tensor = reward_tensor

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "RewardFunctionConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = RewardFunctionConfig(reward_tensor=d["reward_tensor"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation  of the object
        """
        d: Dict[str, Any] = {}
        tensor = list(self.reward_tensor)
        for i in range(len(tensor)):
            if isinstance(tensor[i], np.ndarray):
                tensor[i] = tensor[i].tolist()
        d["reward_tensor"] = list(tensor)
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"reward_tensor:{self.reward_tensor}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "RewardFunctionConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return RewardFunctionConfig.from_dict(json.loads(json_str))



from typing import Dict, Any, Union, List
from csle_base.json_serializable import JSONSerializable
from csle_common.dao.emulation_config.five_g_subscriber_config import FiveGSubscriberConfig
from csle_common.util.general_util import GeneralUtil


class FiveGConfig(JSONSerializable):
    """
    Represents the configuration of the 5G services and managers in a CSLE emulation
    """

    def __init__(self, five_g_core_manager_log_file: str, five_g_core_manager_log_dir: str,
                 five_g_core_manager_max_workers: int,
                 five_g_cu_manager_log_file: str, five_g_cu_manager_log_dir: str,
                 five_g_cu_manager_max_workers: int,
                 five_g_du_manager_log_file: str, five_g_du_manager_log_dir: str,
                 five_g_du_manager_max_workers: int, subscribers: List[FiveGSubscriberConfig],
                 core_backhaul_ip: str,
                 cu_backhaul_ips: List[str], cu_fronthaul_ips: List[str],
                 du_fronthaul_ips: List[str], du_cus: List[str],
                 time_step_len_seconds: int = 15, five_g_core_manager_port: int = 50052,
                 five_g_cu_manager_port: int = 50053, five_g_du_manager_port: int = 50054,
                 version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param version: the version
        :param subscribers: the 5G subscribers
        :param core_backhaul_ip: the IP of the core network as seen to the RAN
        :param cu_backhaul_ips: the backhaul IPs of the CUs
        :param cu_fronthaul_ips: the fronthaul IPs of the CUs
        :param du_fronthaul_ips: the fronthaul IPs of the DUs
        :param du_cus: the CUs for each DU
        :param five_g_core_manager_port: the GRPC port of the 5G core manager
        :param five_g_core_manager_log_file: Log file of the 5G core manager
        :param five_g_core_manager_log_dir: Log dir of the 5G core manager
        :param five_g_core_manager_max_workers: Max GRPC workers of the 5G core manager
        :param five_g_cu_manager_port: the GRPC port of the 5G cu manager
        :param five_g_cu_manager_log_file: Log file of the 5G cu manager
        :param five_g_cu_manager_log_dir: Log dir of the 5G cu manager
        :param five_g_cu_manager_max_workers: Max GRPC workers of the 5G cu manager
        :param five_g_du_manager_port: the GRPC port of the 5G du manager
        :param five_g_du_manager_log_file: Log file of the 5G du manager
        :param five_g_du_manager_log_dir: Log dir of the 5G du manager
        :param five_g_du_manager_max_workers: Max GRPC workers of the 5G du manager
        """
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version

        self.five_g_core_manager_port = five_g_core_manager_port
        self.five_g_core_manager_log_file = five_g_core_manager_log_file
        self.five_g_core_manager_log_dir = five_g_core_manager_log_dir
        self.five_g_core_manager_max_workers = five_g_core_manager_max_workers

        self.five_g_cu_manager_port = five_g_cu_manager_port
        self.five_g_cu_manager_log_file = five_g_cu_manager_log_file
        self.five_g_cu_manager_log_dir = five_g_cu_manager_log_dir
        self.five_g_cu_manager_max_workers = five_g_cu_manager_max_workers

        self.five_g_du_manager_port = five_g_du_manager_port
        self.five_g_du_manager_log_file = five_g_du_manager_log_file
        self.five_g_du_manager_log_dir = five_g_du_manager_log_dir
        self.five_g_du_manager_max_workers = five_g_du_manager_max_workers

        self.subscribers = subscribers
        self.core_backhaul_ip = core_backhaul_ip
        self.cu_backhaul_ips = cu_backhaul_ips
        self.cu_fronthaul_ips = cu_fronthaul_ips
        self.du_fronthaul_ips = du_fronthaul_ips
        self.du_cus = du_cus

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "FiveGConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = FiveGConfig(
            time_step_len_seconds=d["time_step_len_seconds"], version=d["version"],
            five_g_core_manager_log_file=d["five_g_core_manager_log_file"],
            five_g_core_manager_log_dir=d["five_g_core_manager_log_dir"],
            five_g_core_manager_max_workers=d["five_g_core_manager_max_workers"],
            five_g_core_manager_port=d["five_g_core_manager_port"],
            five_g_cu_manager_log_file=d["five_g_cu_manager_log_file"],
            five_g_cu_manager_log_dir=d["five_g_cu_manager_log_dir"],
            five_g_cu_manager_max_workers=d["five_g_cu_manager_max_workers"],
            five_g_cu_manager_port=d["five_g_cu_manager_port"],
            five_g_du_manager_log_file=d["five_g_du_manager_log_file"],
            five_g_du_manager_log_dir=d["five_g_du_manager_log_dir"],
            five_g_du_manager_max_workers=d["five_g_du_manager_max_workers"],
            five_g_du_manager_port=d["five_g_du_manager_port"],
            subscribers=list(map(lambda x: FiveGSubscriberConfig.from_dict(x), d["subscribers"])),
            core_backhaul_ip=d["core_backhaul_ip"],
            cu_backhaul_ips=d["cu_backhaul_ips"],
            cu_fronthaul_ips=d["cu_fronthaul_ips"],
            du_fronthaul_ips=d["du_fronthaul_ips"],
            du_cus=d["du_cus"]
        )
        return obj

    def to_dict(self) -> Dict[str, Union[str, int, List[str], List[Dict[str, Union[str, int]]]]]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int, List[str], List[Dict[str, Union[str, int]]]]] = {}
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["five_g_core_manager_log_file"] = self.five_g_core_manager_log_file
        d["five_g_core_manager_log_dir"] = self.five_g_core_manager_log_dir
        d["five_g_core_manager_max_workers"] = self.five_g_core_manager_max_workers
        d["five_g_core_manager_port"] = self.five_g_core_manager_port
        d["five_g_cu_manager_log_file"] = self.five_g_cu_manager_log_file
        d["five_g_cu_manager_log_dir"] = self.five_g_cu_manager_log_dir
        d["five_g_cu_manager_max_workers"] = self.five_g_cu_manager_max_workers
        d["five_g_cu_manager_port"] = self.five_g_cu_manager_port
        d["five_g_du_manager_log_file"] = self.five_g_du_manager_log_file
        d["five_g_du_manager_log_dir"] = self.five_g_du_manager_log_dir
        d["five_g_du_manager_max_workers"] = self.five_g_du_manager_max_workers
        d["five_g_du_manager_port"] = self.five_g_du_manager_port
        d["subscribers"] = list(map(lambda x: x.to_dict(), self.subscribers))
        d["core_backhaul_ip"] = self.core_backhaul_ip
        d["cu_backhaul_ips"] = self.cu_backhaul_ips
        d["cu_fronthaul_ips"] = self.cu_fronthaul_ips
        d["du_fronthaul_ips"] = self.du_fronthaul_ips
        d["du_cus"] = self.du_cus
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return (f"time_step_len_seconds: {self.time_step_len_seconds}, version: {self.version}, "
                f"five_g_core_manager_log_file: {self.five_g_core_manager_log_file}, "
                f"five_g_core_manager_log_dir: {self.five_g_core_manager_log_dir}, "
                f"five_g_core_manager_max_workers: {self.five_g_core_manager_max_workers}, "
                f"five_g_core_manager_port: {self.five_g_core_manager_port}"
                f"five_g_cu_manager_log_file: {self.five_g_cu_manager_log_file}, "
                f"five_g_cu_manager_log_dir: {self.five_g_cu_manager_log_dir}, "
                f"five_g_cu_manager_max_workers: {self.five_g_cu_manager_max_workers}, "
                f"five_g_cu_manager_port: {self.five_g_cu_manager_port}"
                f"five_g_du_manager_log_file: {self.five_g_du_manager_log_file}, "
                f"five_g_du_manager_log_dir: {self.five_g_du_manager_log_dir}, "
                f"five_g_du_manager_max_workers: {self.five_g_du_manager_max_workers}, "
                f"five_g_du_manager_port: {self.five_g_du_manager_port}, "
                f"subscribers: {self.subscribers}, core_backhaul_ip: {self.core_backhaul_ip},"
                f"cu_backhaul_ips: {self.cu_backhaul_ips}, cu_fronthaul_ips: {self.cu_fronthaul_ips}, "
                f"du_fronthaul_ips: {self.du_fronthaul_ips}, du_cus: {self.du_cus}")

    @staticmethod
    def from_json_file(json_file_path: str) -> "FiveGConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return FiveGConfig.from_dict(json.loads(json_str))

    def copy(self) -> "FiveGConfig":
        """
        :return: a copy of the DTO
        """
        return FiveGConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "FiveGConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.core_backhaul_ip = GeneralUtil.replace_first_octet_of_ip(ip=config.core_backhaul_ip,
                                                                        ip_first_octet=ip_first_octet)
        config.cu_backhaul_ips = list(
            map(lambda x: GeneralUtil.replace_first_octet_of_ip(ip=x, ip_first_octet=ip_first_octet),
                config.cu_backhaul_ips))
        config.cu_fronthaul_ips = list(
            map(lambda x: GeneralUtil.replace_first_octet_of_ip(ip=x, ip_first_octet=ip_first_octet),
                config.cu_fronthaul_ips))
        config.du_fronthaul_ips = list(
            map(lambda x: GeneralUtil.replace_first_octet_of_ip(ip=x, ip_first_octet=ip_first_octet),
                config.du_fronthaul_ips))
        config.du_cus = list(
            map(lambda x: GeneralUtil.replace_first_octet_of_ip(ip=x, ip_first_octet=ip_first_octet),
                config.du_cus))
        return config

    @staticmethod
    def schema() -> "FiveGConfig":
        """
        :return: get the schema of the DTO
        """
        return FiveGConfig(
            version="0.0.1", time_step_len_seconds=15,
            five_g_core_manager_log_file="five_g_core_manager.log",
            five_g_core_manager_port=50052,
            five_g_core_manager_log_dir="/",
            five_g_core_manager_max_workers=10,
            five_g_cu_manager_log_file="five_g_cu_manager.log",
            five_g_cu_manager_port=50052,
            five_g_cu_manager_log_dir="/",
            five_g_cu_manager_max_workers=10,
            five_g_du_manager_log_file="five_g_du_manager.log",
            five_g_du_manager_port=50052,
            five_g_du_manager_log_dir="/",
            five_g_du_manager_max_workers=10,
            subscribers=[], core_backhaul_ip="127.0.0.1",
            cu_backhaul_ips=["127.0.0.1"], cu_fronthaul_ips=["127.0.0.1"],
            du_fronthaul_ips=["127.0.0.1"], du_cus=["127.0.0.1"]
        )



import logging
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil


class FlagsController:
    """
    Class managing flags in the emulation environments
    """

    @staticmethod
    def create_flags(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Connects to a node in the emulation and creates the flags according to a given flags config

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for flags_conf in emulation_env_config.flags_config.node_flag_configs:
            if flags_conf.physical_host_ip != physical_server_ip:
                continue
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                        ip=flags_conf.docker_gw_bridge_ip)

            for flag in flags_conf.flags:
                logger.info(f"Creating flag:{flag.name} on {flags_conf.docker_gw_bridge_ip} ({flags_conf.ip})")
                cmd = constants.COMMANDS.SUDO_RM_RF + " {}".format(flag.path)
                EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                              conn=emulation_env_config.get_connection(
                                                  ip=flags_conf.docker_gw_bridge_ip))
                cmd = constants.COMMANDS.SUDO_TOUCH + " {}".format(flag.path)
                EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                              conn=emulation_env_config.get_connection(
                                                  ip=flags_conf.docker_gw_bridge_ip))
                cmd = constants.COMMANDS.ECHO + " '{}' >> {}".format(flag.name, flag.path)
                EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                              conn=emulation_env_config.get_connection(
                                                  ip=flags_conf.docker_gw_bridge_ip))

            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)



from typing import Dict, Any, Union
from csle_common.dao.simulation_config.state_type import StateType
from csle_base.json_serializable import JSONSerializable


class State(JSONSerializable):
    """
    DTO representing the state of a simulation environment
    """

    def __init__(self, id: int, name: str, descr: str, state_type: StateType):
        """
        Initializes the DTO

        :param id: the id of the state
        :param name: the name of the state
        :param descr: a description of the state
        :param state_type: the state type
        """
        self.id = id
        self.name = name
        self.descr = descr
        self.state_type = state_type

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "State":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = State(id=d["id"], name=d["name"], descr=d["descr"], state_type=StateType(d["state_type"]))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int, StateType]] = {}
        d["id"] = self.id
        d["name"] = self.name
        d["descr"] = self.descr
        d["state_type"] = self.state_type.value
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"id:{self.id}, name:{self.name}, descr:{self.descr}, state_type: {self.state_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "State":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return State.from_dict(json.loads(json_str))



import logging
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil
from csle_common.dao.emulation_config.vulnerability_type import VulnType


class VulnerabilitiesController:
    """
    Class managing vulnerabilities in the emulation environments
    """

    @staticmethod
    def create_vulns(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Creates vulnerabilities in an emulation environment according to a specified vulnerabilities configuration

        :param emulation_env_config: the emulation environment configuration
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        vulnerabilities = emulation_env_config.vuln_config.node_vulnerability_configs
        for vuln in vulnerabilities:
            if vuln.physical_host_ip != physical_server_ip:
                continue
            logger.info(f"Creating vulnerability on ip: {vuln.docker_gw_bridge_ip}, "
                        f"type: {vuln.vuln_type}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=vuln.docker_gw_bridge_ip)

            # Update sudoers file
            if vuln.vuln_type == VulnType.PRIVILEGE_ESCALATION:

                # Restore/Backup sudoers file
                cmd = "sudo cp /etc/sudoers.bak /etc/sudoers"
                EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(
                    ip=vuln.docker_gw_bridge_ip))

                # Install sudoers vulnerability
                if vuln.cve is not None and vuln.cve.lower() == constants.EXPLOIT_VULNERABILITES.CVE_2010_0426:
                    cmd = "sudo su root -c \"echo '{} ALL=NOPASSWD: sudoedit /etc/fstab' >> /etc/sudoers\""
                elif vuln.cve is not None and vuln.cve.lower() == constants.EXPLOIT_VULNERABILITES.CVE_2015_5602:
                    cmd = "sudo su root -c \"echo '{} ALL=NOPASSWD: sudoedit /home/*/*/esc.txt' >> /etc/sudoers\""
                else:
                    raise ValueError("CVE not recognized:{}".format(vuln.cve))

                if vuln.credentials is None:
                    raise ValueError("Could not find credentials corresponding to node vulnerability config.")
                for cr in vuln.credentials:
                    cmd = cmd.format(cr.username)
                    o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                            conn=emulation_env_config.get_connection(
                                                                ip=vuln.docker_gw_bridge_ip))
                    cmd = "sudo chmod 440 /etc/sudoers"
                    o, e, _ = EmulationUtil.execute_ssh_cmd(
                        cmd=cmd, conn=emulation_env_config.get_connection(ip=vuln.docker_gw_bridge_ip))



from typing import List, Union
import docker
from docker.models.containers import Container
import os
import json
from csle_common.dao.docker.docker_container_metadata import DockerContainerMetadata
from csle_common.dao.docker.docker_env_metadata import DockerEnvMetadata
from csle_common.metastore.metastore_facade import MetastoreFacade
import csle_common.constants.constants as constants


class DockerUtil:
    """
    Utility class for extracting information from running emulation environments
    """

    @staticmethod
    def parse_runnning_emulation_infos() -> List[DockerEnvMetadata]:
        """
        Queries docker to get a list of all running emulation environments

        :return: a list of environment DTOs
        """
        client_1: docker.DockerClient = docker.from_env()
        client_2: docker.APIClient = docker.APIClient(base_url=constants.DOCKER.UNIX_DOCKER_SOCK_URL)
        parsed_containers = DockerUtil.parse_running_containers(client_1=client_1, client_2=client_2)
        emulations: List[str] = list(set(list(map(lambda x: x.emulation, filter(lambda x: x is not None,
                                                                                parsed_containers)))))
        parsed_envs = DockerUtil.parse_running_emulation_envs(emulations=emulations, containers=parsed_containers)
        return parsed_envs

    @staticmethod
    def get_container_hex_id(name: str) -> Union[str, None]:
        """
        Queries the docker engine for the id of a container with a given name

        :return: the id
        """
        client_1: docker.DockerClient = docker.from_env()
        client_2: docker.APIClient = docker.APIClient(base_url=constants.DOCKER.UNIX_DOCKER_SOCK_URL)
        containers: List[DockerContainerMetadata] = DockerUtil.parse_running_containers(client_1=client_1,
                                                                                        client_2=client_2)
        for container in containers:
            if container.name == name:
                return container.id
        return None

    @staticmethod
    def parse_running_containers(client_1: docker.DockerClient, client_2: docker.APIClient) \
            -> List[DockerContainerMetadata]:
        """
        Queries docker to get a list of all running containers

        :param client_1: docker client 1
        :param client_2:  docker client 2
        :return: list of parsed running containers
        """
        containers: List[Container] = client_1.containers.list()
        parsed_containers = DockerUtil.parse_containers(containers=containers, client2=client_2)
        return parsed_containers

    @staticmethod
    def parse_stopped_containers(client_1: docker.DockerClient, client2: docker.APIClient) \
            -> List[DockerContainerMetadata]:
        """
        Queries docker to get a list of all stopped csle containers

        :param client_1: docker client 1
        :param client2: docker client 2
        :return: list of parsed containers
        """
        containers: List[docker.models.containers.Container] = client_1.containers.list(all=True)
        stopped_containers: List[docker.models.containers.Container] = list(filter(
            lambda x: (x.status == constants.DOCKER.CONTAINER_EXIT_STATUS
                       or x.status == constants.DOCKER.CONTAINER_CREATED_STATUS), containers))
        parsed_containers = DockerUtil.parse_containers(containers=stopped_containers, client2=client2)
        return parsed_containers

    @staticmethod
    def parse_running_emulation_envs(emulations: List[str], containers: List[DockerContainerMetadata]) \
            -> List[DockerEnvMetadata]:
        """
        Queries docker to get a list of all active emulation environments

        :param emulations: list of csle emulations
        :param containers: list of running csle containers
        :return: list of parsed emulation environments
        """
        parsed_envs = []
        for em in emulations:
            em_containers = list(filter(lambda x: x.emulation == em, containers))
            subnet_prefix = constants.COMMANDS.DOT_DELIM.join(em_containers[0].ip.rsplit(
                constants.COMMANDS.DOT_DELIM)[0:-1])
            subnet_mask = subnet_prefix + constants.COMMANDS.SLASH_DELIM + str(em_containers[0].ip_prefix_len)

            config = None
            em_record = MetastoreFacade.get_emulation_by_name(name=em)
            if em_record is not None:
                config = em_record

            p_env = DockerEnvMetadata(containers=em_containers, name=em, subnet_prefix=subnet_mask,
                                      subnet_mask=subnet_mask, level=em_containers[0].level, config=config,
                                      kafka_config=None)
            parsed_envs.append(p_env)
        return parsed_envs

    @staticmethod
    def parse_containers(containers: List[docker.models.containers.Container], client2: docker.APIClient) \
            -> List[DockerContainerMetadata]:
        """
        Queries docker to get a list of running or stopped csle containers

        :param containers: list of containers to parse
        :param client2: docker client
        :return: List of parsed container DTOs
        """
        parsed_containers = []
        for c in containers:
            if constants.CONTAINER_IMAGES.CSLE_PREFIX in c.name:
                name_parts = c.name.split("-")
                container_name_2 = name_parts[0]
                level = name_parts[1]
                inspect_info = client2.inspect_container(c.id)
                net = None
                if len(list(inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS].keys())) > 0:
                    net = list(inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS].keys())[0]
                labels = c.labels
                config_path = ""
                dir_path = ""
                emulation = ""
                kafka_config = ""
                if constants.DOCKER.CFG in labels:
                    config_path = labels[constants.DOCKER.CFG]
                if constants.DOCKER.CONTAINER_CONFIG_DIR in labels:
                    dir_path = labels[constants.DOCKER.CONTAINER_CONFIG_DIR]
                if constants.DOCKER.EMULATION in labels:
                    emulation = labels[constants.DOCKER.EMULATION]
                if constants.DOCKER.KAFKA_CONFIG in labels:
                    kafka_config = labels[constants.DOCKER.KAFKA_CONFIG]
                ip = ""
                network_id = -1
                gateway = ""
                mac = ""
                ip_prefix_len = 0
                if net is not None:
                    ip = inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS][net][
                        constants.DOCKER.IP_ADDRESS_INFO]
                    network_id = inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS][net][
                        constants.DOCKER.NETWORK_ID_INFO]
                    gateway = inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS][net][
                        constants.DOCKER.GATEWAY_INFO]
                    mac = inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS][net][
                        constants.DOCKER.MAC_ADDRESS_INFO]
                    ip_prefix_len = inspect_info[constants.DOCKER.NETWORK_SETTINGS][constants.DOCKER.NETWORKS][net][
                        constants.DOCKER.IP_PREFIX_LEN_INFO]
                if net is None:
                    net_str = ""
                else:
                    net_str = net
                parsed_c = DockerContainerMetadata(
                    name=c.name, status=c.status, short_id=c.short_id, image_short_id=c.image.short_id,
                    image_tags=c.image.tags, id=c.id,
                    created=inspect_info[constants.DOCKER.CREATED_INFO],
                    ip=ip, network_id=network_id, gateway=gateway, mac=mac, ip_prefix_len=ip_prefix_len,
                    name2=container_name_2, level=level,
                    hostname=inspect_info[constants.DOCKER.CONFIG][constants.DOCKER.HOSTNAME_INFO],
                    image_name=inspect_info[constants.DOCKER.CONFIG][constants.DOCKER.IMAGE],
                    net=net_str, dir=dir_path, config_path=config_path,
                    container_handle=c, emulation=emulation, kafka_container=kafka_config)
                parsed_containers.append(parsed_c)
        return parsed_containers

    @staticmethod
    def get_docker_gw_bridge_ip(container_id: str) -> str:
        """
        Gets the docker gw bridge ip of a container

        :param container_id: the id of the container
        :return: the ip in the gw bridge network
        """
        cmd = constants.DOCKER.INSPECT_DOCKER_GWBRIDGE
        stream = os.popen(cmd)
        json_output = stream.read()
        docker_gw_bridge_info = json.loads(json_output)[0]
        containers = docker_gw_bridge_info[constants.DOCKER.CONTAINERS_KEY]
        if container_id in containers:
            container = containers[container_id]
            ip = container[constants.DOCKER.IPV4_KEY]
            if isinstance(ip, str):
                ip = ip.split("/")[0]
                return ip
        raise ValueError(f"The container with id:{container_id} does not have an IP in the docker gw bridge network. "
                         f"Containers: {containers}")
